<?php

declare(strict_types=1);

namespace Apexl\Vault\Aura\SqlQuery;

use Aura\SqlQuery\Common\QuoterInterface;

class QueryFactory extends \Aura\SqlQuery\QueryFactory
{
    protected function newQuoter(): QuoterInterface
    {
        return new Quoter();
    }

}
